/*
 * pbi_message_parse.c
 *
 *  Created by Brian Quan on 01/11/2012.
 *  Copyright (c) 2012 Brian Quan. All rights reserved.
 */

#include "pbi_message_parse.h"
#include <stdlib.h>
#include <string.h>
#include <CoreFoundation/CoreFoundation.h>

#define HEADER_SIZE_IN_BYTES (8 + 4 + 4 + 1)
#define CRC_SIZE_IN_BYTES 2

typedef struct datastream
{
	const unsigned char *buffer;
	const unsigned char *current;
	unsigned int size;
	int left;
	pbi_status status;
} datastream_t;

void ds_reset(datastream_t* stream)
{
	stream->current = stream->buffer;
	stream->left = stream->size;
	stream->status = kPbi_ok;
}

void ds_error(datastream_t *stream, pbi_status error)
{
	stream->status = error;
}

pbi_status ds_get_status(const datastream_t* stream)
{
	return stream->status;
}

#if __cplusplus
extern "C" {
#endif
    void NSLog(CFStringRef format, ...);
    void NSLogv(CFStringRef format, va_list args);
#if __cplusplus
}
#endif

/**
 * Calculates CRC-XMODEM checksum. The following one is the simplest message possible (header only, no data channel)
 *
 * Data:		24 56 42 50 42 69 24 00 00 00 00 00 00 00 00 00 2C
 * Checksum:	ff 73
 *
 * http://www.lammertbies.nl/comm/info/crc-calculation.html
 */

int ds_verify_checksum(datastream_t* stream)
{
    /*
    int i1 = 0;
    while (i1 < stream->size)
    {
        printf("%02X",(int)stream->buffer[i1]);
        i1++;
    }
    printf("\n");
    */
	pbi_u16 Polynomial = 4129;
	pbi_u16 CRC = 0;
	int left = stream->size - CRC_SIZE_IN_BYTES - 2;
	pbi_u16 expected_checksum = ((pbi_u16)stream->buffer[stream->size - 4]) << 8 | (pbi_u16)stream->buffer[stream->size - 3];
	const pbi_u8* ptr = stream->buffer;
	int i;
	// CRC-XMODEM
	while (left--)
	{
		CRC ^= (((pbi_u16)(*ptr++)) << 8);
		for (i = 0; i < 8; i++)
		{
			if (CRC & 0x8000)
			{
				CRC = (CRC << 1) ^ Polynomial;
			}
			else
			{
				CRC <<= 1;
			}
		}
	}
	if (CRC != expected_checksum)
	{
		ds_error(stream, kPbi_err_checksum);
        printf("-----------------\n");
        printf("Check sum error! bytes size %d\n", stream->size);
		return 0;
	}
	return 1;
}

pbi_status ds_init(datastream_t* stream, const unsigned char* buffer, unsigned int size)
{
	stream->buffer = buffer;
	stream->size = size;
	ds_reset(stream);
	if (buffer == 0 || size == 0)
	{
		ds_error(stream, kPbi_err_invalid_buffer);
	}
	else if (size < (HEADER_SIZE_IN_BYTES + CRC_SIZE_IN_BYTES))
	{
		ds_error(stream, kPbi_err_header_too_short);
	}
	else
	{
		ds_verify_checksum(stream);
	}
	return ds_get_status(stream);
}

void ds_advance(datastream_t *stream, int count)
{
	if (count > stream->left)
	{
		stream->current = stream->buffer + stream->size;
		stream->left = 0;
		ds_error(stream, kPbi_err_unexpected_eos);
	}
	else
	{
		stream->current += count;
		stream->left -= count;
	}
}

int ds_get_chars(datastream_t *stream, pbi_c* dest, int count)
{
	if (stream->left < count)
	{
		memset(dest, 0, count);
		count = stream->left;
		ds_error(stream, kPbi_err_unexpected_eos);
	}
	memcpy(dest, stream->current, count);
	ds_advance(stream, count);
	return count;
}

int ds_get_u8_v(datastream_t *stream, pbi_u8 *dest, int count)
{
	return ds_get_chars(stream, (pbi_c*)dest, count);
}

pbi_c ds_get_c(datastream_t *stream)
{
	pbi_c result;
	ds_get_chars(stream, &result, 1);
	return result;
}

pbi_u8 ds_get_u8(datastream_t *stream)
{
	pbi_u8 result;
	ds_get_u8_v(stream, &result, 1);
	return result;
}

pbi_u16 ds_get_u16(datastream_t *stream)
{
	pbi_u8 bytes[2];
	ds_get_u8_v(stream, bytes, 2);
	return (((pbi_u16)bytes[0]) << 8) | ((pbi_u16)bytes[1]);
}

pbi_u24 ds_get_u24(datastream_t *stream)
{
	pbi_u8 bytes[3];
	ds_get_u8_v(stream, bytes, 3);
	return (((pbi_u24)bytes[0]) << 16) | (((pbi_u24)bytes[1]) << 8) | ((pbi_u24)bytes[2]);
}

pbi_u32 ds_get_u32(datastream_t *stream)
{
	pbi_u8 bytes[4];
	ds_get_u8_v(stream, bytes, 4);
	return (((pbi_u32)bytes[0]) << 24) | (((pbi_u32)bytes[1]) << 16) | (((pbi_u32)bytes[2]) << 8) | ((pbi_u32)bytes[3]);
}

pbi_s16 ds_get_s16(datastream_t *stream)
{
	return (pbi_s16)ds_get_u16(stream);
}

pbi_u24 ds_get_s24(datastream_t *stream)
{
	pbi_s24 result = (pbi_s24)ds_get_u24(stream);
	if (result & 0x00800000)
	{
		result |= 0xff000000; // pastes sign bit to fill
	}
	return result;
}

pbi_s32 ds_get_s32(datastream_t *stream)
{
	return (pbi_s32)ds_get_u32(stream);
}

pbi_status ds_get_header(datastream_t *stream, pbi_header_t *header)
{
	memset(header, 0, sizeof(pbi_header_t));
	ds_get_chars(stream, &header->header[0], sizeof(header->header));
	header->standard_channel_flags = ds_get_u32(stream);
	header->extended_channel_flags = ds_get_u32(stream);
	header->comma = ds_get_c(stream);
	if (strncmp(header->header, "$VBSPT$", 7) || header->comma != ',')
	{
		ds_error(stream, kPbi_err_data_format);
	}
	return ds_get_status(stream);
}

pbi_status ds_get_sc_fields(datastream_t *stream, pbi_u32 sc_flags, pbi_sc_fields_t* sc_fields)
{
	memset(sc_fields, 0, sizeof(pbi_sc_fields_t));
	if (sc_flags & kSCFlags_Satellites)
	{
		pbi_u8 sats = ds_get_u8(stream);
		sc_fields->num_satellites = (sats & 0x7f);
		sc_fields->using_dgps = !!(sats & 0x80);
	}
	if (sc_flags & kSCFlags_Time)
	{
		sc_fields->timestamp = ds_get_u24(stream) * 10L;
	}
	if (sc_flags & kSCFlags_Latitude)
	{
		sc_fields->latitude = ((pbi_scalar)ds_get_s32(stream)) / 6000000.0;
	}
	if (sc_flags & kSCFlags_Longitude)
	{
		sc_fields->longitude = ((pbi_scalar)ds_get_s32(stream)) / -6000000.0;
	}
	if (sc_flags & kSCFlags_Speed)
	{
		sc_fields->speed = ((pbi_scalar)ds_get_u16(stream)) / 100.0;
	}
	if (sc_flags & kSCFlags_Heading)
	{
		sc_fields->heading = ((pbi_scalar)ds_get_u16(stream)) / 100.0;
	}
	if (sc_flags & kSCFlags_Height)
	{
		sc_fields->height = ((pbi_scalar)ds_get_s24(stream)) / 100.0;
	}
	if (sc_flags & kSCFlags_VerticalSpeed)
	{
		sc_fields->vertical_speed = (pbi_scalar)ds_get_s16(stream);
	}
	if (sc_flags & kSCFlags_LongitudinalAcceleration)
	{
		sc_fields->longitudinal_acceleration = ((pbi_scalar)ds_get_s16(stream)) / 100.0;
	}
	if (sc_flags & kSCFlags_LateralAcceleration)
	{
		sc_fields->lateral_acceleration = ((pbi_scalar)ds_get_s16(stream)) / 100.0;
	}
	if (sc_flags & kSCFlags_Distance)
	{
		sc_fields->distance = ((pbi_scalar)ds_get_u32(stream)) / 128000.0;
	}
	if (sc_flags & kSCFlags_Temperature)
	{
		sc_fields->temperature = ((pbi_scalar)ds_get_u32(stream)); // 32 bit?
	}
	if (sc_flags & kSCFlags_MediaFreeSpace)
	{
		pbi_s24 mfs = ds_get_u24(stream);
		sc_fields->media_free_space = ((pbi_scalar)(mfs - 0xef7ff)) / (pbi_scalar)0xef7ff * 100.0;
	}
	if (sc_flags & kSCFlags_BatteryVoltage)
	{
		sc_fields->battery_voltage = ds_get_u16(stream);
	}
	return ds_get_status(stream);
}

pbi_status ds_get_ec_fields(datastream_t *stream, pbi_u32 ec_flags, pbi_ec_fields_t* ec_fields)
{
	memset(ec_fields, 0, sizeof(pbi_ec_fields_t));
	if (ec_flags & kECFlags_BatteryTimeToEmpty)
	{
		ec_fields->battery_time_to_empty = ds_get_u16(stream);
	}
	if (ec_flags & kECFlags_BatteryVoltage)
	{
		ec_fields->battery_voltage = ds_get_u16(stream);
	}
	if (ec_flags & kECFlags_BatteryChargeWhenFull)
	{
		ec_fields->battery_charge_when_full = ds_get_u16(stream);
	}
	if (ec_flags & kECFlags_BatteryCurrentCharge)
	{
		ec_fields->battery_current_charge = ds_get_u16(stream);
	}
	if (ec_flags & kECFlags_MediaCapacity)
	{
		ec_fields->media_capacity = ds_get_u32(stream);
	}
	if (ec_flags & kECFlags_MediaFreeSpace)
	{
		ec_fields->media_free_space = ds_get_u32(stream);
	}
	if (ec_flags & kECFlags_HDOP)
	{
		ec_fields->HDOP = (pbi_scalar)ds_get_u16(stream) / 100.0;
	}
	return ds_get_status(stream);
}

pbi_status pbi_message_parse(pbi_data_message_t* dest, const unsigned char* raw_data, unsigned int size)
{
	memset(dest, 0, sizeof(pbi_data_message_t));
	datastream_t stream;
    
	(void)((kPbi_ok == ds_init(&stream, raw_data, size)) && (kPbi_ok == ds_get_header(&stream, &dest->header))
	        && (kPbi_ok == ds_get_sc_fields(&stream, dest->header.standard_channel_flags, &dest->sc_fields))
	        && (kPbi_ok == ds_get_ec_fields(&stream, dest->header.extended_channel_flags, &dest->ec_fields)));
	return ds_get_status(&stream);
}

int pbi_message_has_sc_field(const pbi_data_message_t* message, pbi_u32 sc_flags)
{
	return (message->header.standard_channel_flags & sc_flags) == sc_flags;
}

int pbi_message_has_ec_field(const pbi_data_message_t* message, pbi_u32 ec_flags)
{
	return (message->header.extended_channel_flags & ec_flags) == ec_flags;
}

const char * pbi_error_message(pbi_status status)
{
	static struct _pbi_error_msg
	{
		pbi_status code;
		const char* message;
	} s_pbi_error_msg[] = { {kPbi_ok, "kPbi_ok"}, {kPbi_err_checksum, "kPbi_err_checksum"}, {kPbi_err_header_too_short, "kPbi_err_header_too_short"}, {
	        kPbi_err_unexpected_eos, "kPbi_err_unexpected_eos"}, {kPbi_err_data_format, "kPbi_err_data_format"}, {kPbi_err_invalid_buffer,
	        "kPbi_err_invalid_buffer"}, {0, NULL}};
	struct _pbi_error_msg *msg = s_pbi_error_msg;
	while (msg->message != NULL)
	{
		if (msg->code == status)
			return msg->message;
		msg++;
	}
	return NULL;
}
